# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .perplexity import Perplexity
from .chunk import ChunkEvaluator
from .bleu import BLEU, BLEUForDuReader
from .rouge import RougeL, RougeLForDuReader, RougeN, Rouge1, Rouge2
from .glue import AccuracyAndF1, Mcc, PearsonAndSpearman, MultiLabelsMetric
from .distinct import Distinct
from .sighan import DetectionF1, CorrectionF1
from .span import SpanEvaluator
